'''Example scripts.

eigenimages:
    Principal Component Analysis (PCA) demos for 2D images.

pca:
    PCA demos for general data.

buckling_evp:
    Computes buckling load factors and modes.

core_solver.py:
    Basic core solver usage demo.

sparse_evp.py:
    Computes several eigenvalues and eigenvectors of a real symmetric matrix.

sparse_evp_all_options.py:
    Same as sparse_evp.py but with more options.

truncated_svd.py:
    Computes truncated SVD.
'''
